// Copyright 2017 The Crashpad Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "util/posix/scoped_dir.h"

#include "base/logging.h"
#include "base/posix/eintr_wrapper.h"

namespace crashpad {
namespace internal {

void ScopedDIRCloseTraits::Free(DIR* dir) {
  if (dir && IGNORE_EINTR(closedir(dir)) != 0) {
    PLOG(ERROR) << "closedir";
  }
}

}  // namespace internal
}  // namespace crashpad
